package edu.berkeley.qfat.geom;
import javax.media.opengl.*;
import java.util.*;
import com.infomatiq.jsi.*;
import com.infomatiq.jsi.rtree.*;

/** wrapper around the <tt>com.infomatiq.jsi.rtree.RTree</tt> class */
public class RTree<V extends HasBoundingBox> implements Iterable<V> {

    private com.infomatiq.jsi.rtree.RTree rtree =
        new com.infomatiq.jsi.rtree.RTree();

    private int lowid = 0;
    private HashMap<Integer, V> idToV = new HashMap<Integer, V>();
    private HashMap<V, Integer> vToId = new HashMap<V, Integer>();

    public Iterator<V> iterator() { return vToId.keySet().iterator(); }

    private final MyIntProcedure myIntProcedure    = new MyIntProcedure();
    private final com.infomatiq.jsi.Rectangle rect = new com.infomatiq.jsi.Rectangle(0,0,0,0,0,0);
    private final com.infomatiq.jsi.Point point    = new com.infomatiq.jsi.Point(0,0,0);
    private V          found   = null;
    private Visitor<V> visitor = null;

    private static final Properties props = new Properties();
    static {
        props.put("MinNodeEntries", "1");
        props.put("MaxNodeEntries", "5");
    }

    public RTree() { clear(); }

    public void clear() {
        idToV.clear();
        vToId.clear();
        rtree.init(props);
    }

    public void add(V v) { insert(v); }
    public void insert(V v) {
        int id = lowid++;
        idToV.put(id, v);
        vToId.put(v, id);
        rect.set(v.getMinX(), v.getMinY(), v.getMinZ(), v.getMaxX(), v.getMaxY(), v.getMaxZ());
        rtree.add(rect, id);
    }

    public void remove(V v) {
        Integer idi = vToId.get(v);
        if (idi==null) return;
        int id = idi;
        idToV.remove(id);
        vToId.remove(v);
        rect.set(v.getMinX(), v.getMinY(), v.getMinZ(), v.getMaxX(), v.getMaxY(), v.getMaxZ());
        rtree.delete(rect, id);
    }

    public V nearest(Point p) {
        point.set(p.x, p.y, p.z);
        rtree.nearest(point, myIntProcedure, Float.POSITIVE_INFINITY);
        V ret = found;
        found = null;
        return ret;
    }

    public void range(HasBoundingBox v, Visitor vis) {
        visitor = vis;
        rtree.intersects(rect, myIntProcedure);
        visitor = null;
    }

    private class MyIntProcedure implements IntProcedure {
        public boolean execute(int id) {
            if (visitor != null) {
                V v = idToV.get(id);
                visitor.visit(v);
                return true;
            } else {
                found = idToV.get(id);
                return false;
            }
        }
    }
}
